/*
 * Component:  utils
 *
 * Purpose:
 *    Logging implementation using syslog
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 *
 */
#ifndef IMDB_LOGGER_H_INCLUDED
#define IMDB_LOGGER_H_INCLUDED

#include <stdarg.h>
#include <string>
#include <map>
#include <vector>
using namespace std;

#include "IMLogger.h"
#include "dbim.h"

/**
 * Encodes string values to be added to oper log entries
 *
 * @param value  the value to be encoded
 * @return the log-encoded value
 */
string logValue( const char * value );

/**
 * Helper class to collect study level values to be added to oper log entries from database
 */
class StudyLogInfo {
	private:
		string m_study;
		string m_patientName;
		string m_accessionNumber;
		string m_modality;
		string m_patientID;
		string m_studyDate;
		string m_institution;
		string m_studyDescription;
		string m_bodypart;

	public:

	/**
	 * Constructor for creating a new Study Log Info object
	 * @param studyUID  study instance UID
	 */
	StudyLogInfo( const char * studyUID );

	/**
	 * Destructor to destroy this Study Log Info object
	 */
	~StudyLogInfo();

	/**
	 * @return  Log-encoded patient name
	 */
	string getLogPatientName() const {
		return logValue( m_patientName.c_str() );
	}

	/**
	 * @return  Log-encoded accession number
	 */
	string getLogAccessionNumber() const {
		return logValue( m_accessionNumber.c_str() );
	}

	/**
	 * @return  Log-encoded modality
	 */
	string getLogModality() const {
		return logValue( m_modality.c_str() );
	}

	/**
	 * @return  Log-encoded patient ID
	 */
	string getLogPatientID() const {
		return logValue( m_patientID.c_str() );
	}

	/**
	 * @return  Log-encoded study date
	 */
	string getLogStudyDate() const {
		return logValue( m_studyDate.c_str() );
	}

	/**
	 * @return  Log-encoded institution
	 */
	string getLogInstitution() const {
		return logValue( m_institution.c_str() );
	}

	/**
	 * @return  Log-encoded study description
	 */
	string getLogStudyDescription() const {
		return logValue( m_studyDescription.c_str() );
	}

	/**
	 * @return  Log-encoded body part
	 */
	string getLogBodypart() const {
		return logValue( m_bodypart.c_str() );
	}

	/**
	 * @return  Log-encoded study instance UID
	 */
	string getLogStudyUID() const {
		return logValue( m_study.c_str() );
	}

	/**
	 * Fills params with study information to be added to a STUDY/OBJECT level oper log entry.
	 * @param params  output parameter to add study log information to
	 */
	void fillParams( map<string, string> & params ) const;
};


/**
 * Wrapper around a log field
 */
class LogField {
		/** Database field name for fields to be added to database log entries as well */
		string m_dbname;
		/** Value to be logged */
		string m_value;

	public:

		/**
		 * Default constructor creating an empty log field
		 */
		LogField();

		/**
		 * Copy constructor to create a new log field by copying data from another log field object
		 * @param logField original log gield to be copied
		 */
		LogField( const LogField & logField );

		/**
		 * Destructor to destroy the log field
		 */
		~LogField();

		/**
		 * Assignment operator to copy data from another log field object to this log field object
		 * @param logField
		 * @return reference to this log field object
		 */
		LogField & operator= ( const LogField & logField );

		/**
		 * @return Database field name of the field
		 */
		inline const string & getDbname() const {
		    return m_dbname;
		}

		/**
		 * @return  value to be logged
		 */
		inline const string & getValue() const {
		    return m_value;
		}

		/**
		 * Sets database field name
		 * @param dbname new database field name of the field (empty for fields having no corresponding database field)
		 */
		inline void setDbname( const string & dbname ) {
		    m_dbname = dbname;
		}

		/**
		 * @param value new value for this log field
		 */
		inline void setValue( const string & value ) {
		   m_value = value;
		}
};

/**
 * Class for creating oper_info and oper_error log entries and also log them to the log database table
 */
class IMDBLogger : public IMLogger {

	public:

		/**
		 * Create new Database Logger object with logging tag
		 *
		 * @param tag  Logger tag to be used in log messages
		 */
		IMDBLogger( const char * tag ) : IMLogger( tag ) {
		}

		/**
		 * Create log entry in oper_info.log
		 * Log entry will be parsed and added to the log database if needed
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void operInfoLog( const char * mask, ... );

		/**
		 * Create log entry in oper_error.log
		 * Log entry will be parsed and added to the log database if needed
		 *
		 * @param mask   Log message mask (see printf)
		 * @param ...    Arguments to be replaced in mask
		 */
		void operErrorLog( const char * mask, ... );

		/**
		 * Create an oper_info or oper_error log entry using the provided fields.
		 * A log database entry will also be created if the entry intended to be
		 * logged to the database (ie. the level is neither OBJECT nor SERIES).
		 *
		 * fields should contain empty log fields at positions where no log value is specified.
		 *
		 * @param mode    log mode (INFO/ERROR)
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 * @param fields  fields to be logged
		 */
		void operLog( LOG_MODE mode, const char * level, const char * action, const vector<LogField> & fields );

		/**
		 * Create an oper_info or oper_error log entry using the provided parameters.
		 * A log database entry will also be created if the entry intended to be
		 * logged to the database (ie. the level is neither OBJECT nor SERIES).
		 *
		 * Parameters intended to be a map of <code>log fields</code> to <code>log values</code>.
		 *
		 * Log field can be either<UL>
		 * <LI> an SQL name of a log database field (only a subset is supported)</LI>
		 * <LI> a number: index of the field in the log file (index starts at 3 - USERID)</LI>
		 * </UL>
		 *
		 * Log value is the value to be logged for the specified field (unescaped)
		 *
		 * @param mode    log mode (INFO/ERROR)
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 * @param params  paremeters holding values to be logged
		 */
		void operLog( LOG_MODE mode, const char * level, const char * action, const map<string,string> & params );

	private:
		/**
		 * Log into database
		 *
		 * @param mask		Log message mask (see printf)
		 * @param args		Arguments to be replaced in mask
		 */
		void dbLog( const char * mask, va_list args );

		/**
		 * Converts log parameters to a vector of log fields
		 * @param params  paremeters holding values to be logged
		 * @param fields  output parameter - the list of log fields generated from log parameters
		 */
		void params2vector( const map<string, string> & params, vector<LogField> & fields );

		/**
		 * Adds the generated log entry to the oper_info/oper_error log file.
		 * Intended to be called by public operLog() methods only
		 * @param mode   log mode (INFO/ERROR)
		 * @param mask   log message mask (see printf)
		 */
		void operLog( LOG_MODE mode, const char * mask, ... );

		/**
		 * Checks if callback is registered for the given log level/log action and starts a Callback Hook task
		 * if needed.
		 *
		 * @param level   log level (STUDY/SERIES/OBJECT/ADMIN/...)
		 * @param action  log action
		 */
		void startCallbackTask( const char * level, const char * action );
};


#endif // IM_LOGGER_H_INCLUDED
