/*
 * Component:  utils
 *
 * Purpose:
 *   Header file for reference counter implementation
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$ 
 * State            : $State$
 *
 */

#ifndef REFLOCKENTRY_H
#define REFLOCKENTRY_H

#include <stdio.h>

#include <dcmtk/ofstd/oflist.h>
#include <dcmtk/ofstd/ofstring.h>

#include "IMSemaphore.h"

class ReleaseConfig;

/**
* Reference counter implementation. A reference counter may contain several
* referrer entries for an object. Each referrer may refer the object several times.
* Reference counters are stored in a file. If there is no reference on the object the
* reference counter file will be removed.
*
* @note For simple manipulation of reference counters from C/C++ programs you might want to use RefCount objects.
*/
class ReferenceCounter
{
private:

	/** Identifies the object to be referenced */
	OFString m_id;

	/** Create/store the reference counter file in this directory */
	OFString m_lockDir;

	/** Type of the object to be referenced. */
	OFString m_type;

	/** Comment might be added to the reference counter. */
	OFString m_comment;

	/** Semaphore object to lock adding/removing referrers */
   IMSemaphore m_semaphore;

   /** List of referrers */
   OFList<OFString> m_referrers;

   /** Configuration file of callback of types to run command when object is released */
   OFString m_configFile;

   /** Configuration of commands to be triggered when a reference counter is released */
   ReleaseConfig & m_releaseCfg;

   /** true indicates the reference counter object uses a single level */
   bool m_isSingleLevel;

   /**
   * Reads referrers from a reference counter file
   *
   * @filename name of the reference counter file to be read
   *
   * @return true  if and only if the reference counter file was successfully read
	*         false if the reference counter file does not exist
	*
	* @throws IMIOException if an IO error occured
   */
	bool read(const OFString & filename);


   /**
   * Writes referrers to a file stream
   *
   * @file   file stream to write referrers into
   *
   */
	void write(FILE * file);

   /**
   * Writes referrers to a reference counter file
   *
   * @filename name of the reference counter file to be written
   *
   * @return true if and only if referrers was successfully read to the reference counter file
	*
	* @throws IMIOException if an IO error occured
   */
   bool write(const OFString & filename);

	/**
	* Checks if the directory exists and create it. Non existing parent directories will also be created recursively
	*
	* @param dirname   full path of the directory to be created
	*
	* @return true  if the directory exists or the directory was successfully created
	*         false otherwise
	*
	* @throw IOException if the directory or one of the parent directories cannot be created
	*/
	bool checkAndCreateDirectories(const OFString & dirname);

	/**
	* Removes the directory if empty. Also removes empty parent directories recursively under prefix.
	*
	* @param dirname  relative path of the directory to the prefix to be removed
	* @param prefix   remove empty directories only under this path
	*/
	void removeEmptyDirectories(const OFString & prefix, const OFString & dirname);


	/**
	 * Checks and creates hash directories for the reference counter file
	 *
	 * @return true if the hash directories already exists or has successfully been created
	 */
	bool checkAndCreateHashDirs();

	/**
	 * Removes empty hash directories after a reference counter file get deleted
	 *
	 * @return true if the empty hash directories could be successfully deleted
	 */
	bool removeHashDirs();

public:

	/**
	* Constructor for creating a reference counter object
	*
	* @param id          identifies the object to be referenced
	* @param type        type of the object to be referenced (eg. STUDY, OBJECT)
	* @param lockDir     directory where reference counter file will be created
	*                    (if empty tries LOCK_VAR or if it is NULL then the current directory)
	* @param releaseCfg  Configuration containing commands triggered by releasing an object
	*/
	ReferenceCounter(const OFString & id, const OFString & type, const OFString & lockDir, ReleaseConfig & releaseCfg);

	/**
	* Constructor for creating a reference counter object
	*
	* @param id          identifies the object to be referenced
	* @param type        type of the object to be referenced (eg. STUDY, OBJECT)
	* @param releaseCfg  Configuration containing commands triggered by releasing an object
	*/
	ReferenceCounter(const OFString & id, const OFString & type, ReleaseConfig & releaseCfg);

	/**
	* Destructor for destroying a reference counter object
	*/
	~ReferenceCounter();

	/**
	* Retrieves the type of the object
	*
	* @return type of the object
	*/
	inline const OFString & getType() { return m_type; }

	/**
	 * Retrieves the comment added to the reference counter object
	 *
	 * @return comment added to the reference counter object
	 */
	inline const OFString & getComment() { return m_comment; }
	
	/**
	 * Retrieves the list of references on the reference counter object
	 *
	 * @return the list of references on the reference counter object
	 */
	inline const OFList<OFString> & getReferrers() { return m_referrers; }

	/**
	 * Adds comment to the reference counter object or changes existing comment
	 *
	 * @param comment added to the reference counter object
	 */
	inline void setComment(const char * comment) { m_comment = comment; }
	
   /**
   * Creates a new referrer entry for a referenceID.
   *
	* @param referrerID  identifies the referrer entry to be created
   *
   * @return true if and only if the referrer entry is successfully added
   */
	bool addReference(const OFString & referrerId);

	/**
	* Removes a referrer entry from the reference counter list.
	*
	* If there is no more reference on the object then removes reference file and calls onRelease().
   *
	* @param referrerID  identifies the referrer entry to be removed
   *
   * @return true if and only if the referrer entry was successfully removed (and existed before)
	*/
	bool removeReference(const OFString & referrerId);

	/**
	* Checks if the reference counter contains a referrer entry.
	*
	* @param referrerID  identifies the referrer entry to be checked
	*
	* @return true if and only if the reference counter list contains the referrerID.
	*/
	bool containsReference(const OFString & referrerId);

	/**
	* Prints the reference list to the standard output
	*/
	void print();

	/**
	 * Prints only the references to stdout
	 *
	 * @param forJava  if true, also print the comment in the first line in '#COMMENT: comment' format
	 */
	void printRefs(bool forJava = false);

	/**
	 * Prints the entire reference file to stdout
	 */
	void printRaw();


	/**
	 * Tests whether the reference counter file exists.
	 * @return true if and only if the reference counter file exists; false otherwise
	 */
	bool exists();


	/**
	* Loads the reference counter list from file into memory.
	*
	* @return true  if and only if reference counter list was successfully loaded
	*         false if the reference counter file does not exist
	*
	* @throws IMTimeoutException - if the timeout is reached
	*         IMIOException if semaphore file did not exist and could not be created.
	*/	
	bool load();

	/**
	* Saves reference counter list into.
	* If the reference counter list is empty (ie there is no reference entry in it)
	* the existing reference counter list file will be removed.
	*
	* @return true  if and only if reference counter list was successfully loaded
	*/
	bool save();

	/**
	* Checks if callback commands would be started for this object type, and start commands.
	* Callback commands will be read from configuration file.
	*
	* onRelease() will be called when all references are removed from the object
	*/
	void onRelease();
};



/**
* Entries in the configuration file of registered commands to be started when an object is released.
*/
class ReleaseConfigEntry
{
private:
	/** Object type */
	OFString m_type;

	/** Event (RELEASED) */
	OFString m_event;

	/** Command to be executed */
	OFString m_command;
	
public:
	/**
	* Creates a new configuration file entry
	*
	* @param type     Object type
	* @param event    Event (RELEASED)
	* @param command  Command to be executed
	*/
	ReleaseConfigEntry(const OFString & type, const OFString & event, const OFString & command);

	/**
	* Destructor for destroying a ReleaseConfig object
	*/
	~ReleaseConfigEntry();

	/**
	* Checks if this configuration entry is matching to a type
	*
	* @param type  check configuration entry against this type
	*
	* @return true if and only if this configuration entry is matching to the type 
	*/
	inline bool isMatchingType(const OFString & type) {
		return m_type.compare(type) == 0;
	}

	/**
	* Checks if this configuration entry is matching to an event
	*
	* @param event    check configuration entry against this event
	*
	* @return true if and only if this configuration entry is matching to the event
	*/
	inline bool isMatchingEvent(const OFString & event) {
		return m_event.compare(event) == 0;
	}

	/**
	* Checks if this configuration entry is matching to a type and an event
	*
	* @param type   check configuration entry against this type
	* @param event  check configuration entry against this event
	*
	* @return true if and only if this configuration entry is matching to both the type and the event
	*/
	inline bool isMatching(const OFString & type, const OFString & event) {
		return isMatchingType(type) && isMatchingEvent(event);
	}

	/**
	* Retrieves the command to be executed.
	*
	* @return  the registered command to be executed
	*/
	inline const OFString & getCommand() { return m_command; }
};


/**
* Reads configuration file of registered commands to be started when an object is released.
*/
class ReleaseConfig
{
private:

	/** List of configuration file entries */
	OFList<ReleaseConfigEntry *> m_entryList;

public:

	/**
	* Constructor for creating a new ReleaseConfig object
	*
	* @param configFile    name of the configuration file (if empty tries REFERENCE_RELEASE_CONFIG)
	*/
	ReleaseConfig();
	
	/**
	* Destructor for destroying a ReleaseConfig object
	*/
	~ReleaseConfig();

	/**
	* Reads the configuration file
	*/
	void readConfigFile(const OFString & configFile);

	/**
	* Checks if there is any command is registered for this type and starts
	* the registered commands. id will be passed to the command
	*
	* @param type   check commands registered for this object type
	* @param id     ID of the object. Will be passed to the commands.
	*/
	void checkAndStartCommandsFor(const OFString & type, const OFString & event, const OFString & id);
};


#endif
